// ----------------------------------------------------------------------------
// * Raima Database Manager
// *
// * Copyright (c) 2012 Raima Inc.,  All rights reserved.
// *
// * Use of this software, whether in source code format, or in executable,
// * binary object code form, is governed by the Raima LICENSE which
// * is fully described in the LICENSE.TXT file, included within this
// * distribution of files.
// * ----------------------------------------------------------------------------

#include <iostream>

// rdm.h is required in all files using the RDM Database
#include "rdm.h"

// obj15_db.h was generated from the DDL file obj15_db.ddl.
#include "obj15_db.h"

// obj15_db_api.h was generated from the DDL file obj15_db.ddl. with 
// the -cpp option
#include "obj15_db_api.h"

static uint8_t quit = 0;

// Name space for the RDM C++ interface
using namespace RDM_CPP;

// EXAMPLE - obj15 for C++
// This example shows multi-user access to a database using separate threads
// and a remote tfserver.  The connections will use TCP/IP to communicate
// with the tfserver to do their work.  After starting the tfserver, 
// two writer threads are created.  The first begins inserting values into
// the table.  The second waits until the table contains more than 100 rows
// and then begins deleting the oldest.  A reader thread is also created
// that prints out the total number of records, the max value, and the min
// value.  Each thread uses regular locking.

// The DDL (Database Definition Language) is in obj15_db.ddl. The database
// schema is embedded into the application and the database will be created
// the first time the application is run.  Subsequent runs will use the now
// existing database.

// The following RDM C++ methods are illustrated:
//      Db->Open
//      Db->Initialize
//      Db->BeginUpdate
//      Db->End
//      Db->Abort
//      Db->BeginRead
//      Db->New_recordWithFieldValues
//      Db->Get_recordsByKey
//      Cursor->AfterLast
//      Cursor->GetFieldValues
//      Cursor++
// They are fully documented in http://docs.raima.com/rdm/.

// Each of these methods can throw and an exception of type rdm_exception

// The actual database is stored in a directory named 'obj15_db' in this
// project directory.

class obj15
{
public:
    Db_obj15_db    db;

    void InsertData () throw (rdm_exception&);
    void DeleteData() throw (rdm_exception&);
    void DisplayData() throw (rdm_exception&);
    void Initialize (uint16_t init_db=0) throw (rdm_exception&);
    void Cleanup () throw (rdm_exception&);
};

void obj15::Initialize(
    uint16_t init_db) throw (rdm_exception&)
{
    // Open the database and initialize it
    try
    {
        db = db.Open("obj15_db");
        if(init_db)
        {
            db.Initialize();
        }
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") opening/initializing obj15_db database" << endl;
        throw (e);
        return;
    }
    return;
}

void obj15::InsertData() throw (rdm_exception&)
{
    int32_t      ii;
    struct rec01 rec01_rec;
    recordType rtypes[]= {Cursor_rec01::Type (), Cursor::NoMoreTypes()};
        
    try
    {
        // Loop until there is an error or we are told to quit
        for (ii=0; !quit; ii++)
        {
            db.BeginUpdate(rtypes);

            rec01_rec.field01 = ii;
            db.New_rec01_recordWithFieldValues(&rec01_rec);

            db.End();
        }
    }
    catch (rdm_exception& e)
    {
        db.Abort();
        cerr << "Error " << e.error() << " (" << e.what() << ") adding records" << endl;
        throw (e);
        return;
    }
    return;
}

void obj15::DeleteData() throw (rdm_exception&)
{
    int32_t    ii;
    uint64_t   rec_count;
    recordType rtypes[]= {Cursor_rec01::Type (), Cursor::NoMoreTypes()};
        
    try
    {
        // Loop until there is an error or we are told to quit
        for (ii=0; !quit; ii++)
        {
            db.BeginRead(rtypes);
            rec_count = db.Get_rec01_records().GetCount();
            db.End();

            if(rec_count > 100)
            {
                db.BeginUpdate(rtypes);
                db.Get_rec01_recordsBy_field01().First().Delete();
                db.End();
            }
            else
            {
                /* Wait for a bit to allow the inserter thread to get more records
                 * in the database
                 */
                sleep(1);
            }
        }
    }
    catch (rdm_exception& e)
    {
        db.Abort();
        cerr << "Error " << e.error() << " (" << e.what() << ") deleting records" << endl;
        throw (e);
        return;
    }
    return;
}
    
void obj15::DisplayData() throw (rdm_exception&)
{
    Cursor_rec01 rec01_cursor;
    uint64_t     rec_count = 0;
    int32_t      field01_min;
    int32_t      field01_max;
    char         msg[256];
    recordType   rtypes[]= {Cursor_rec01::Type (), Cursor::NoMoreTypes()};

    cout << endl;
    cout << "Total Records   Min Val  Max Val" << endl;
    cout << "-------------   -------  -------" << endl;

    try
    {
        // Loop until there is an error or we are told to quit
        while(!quit)
        {
            // Wait one second
            sleep(1);

            db.BeginRead(rtypes);

            rec01_cursor = db.Get_rec01_recordsBy_field01();

            rec_count = rec01_cursor.GetCount();

            rec01_cursor.First();
            rec01_cursor.Get_field01(&field01_min);

            rec01_cursor.Last();
            rec01_cursor.Get_field01(&field01_max);

            db.End();

            sprintf(msg, "       %6lu    %6d   %6d", rec_count, field01_min, field01_max);
            cout << msg << endl;
        }
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") reading records" << endl;
        if(e.error() != S_UNAVAIL)
        {
            db.End();
        }
        return;
    }   
}

void obj15::Cleanup() throw (rdm_exception&)
{
    db.Release();
    return;
}

/* Start an external TFS Process */
static pid_t start_tfs_process(void)
{
    pid_t pid = fork();
    if (pid < 0)
    {
        cerr << "Error forking (" << strerror(errno) << ")" << endl;
        exit(EXIT_FAILURE);
    }

    if (pid == 0)
    {
        int32_t     stat;
        const char *tfserver_main_argv[] = {
            "tfserver"
        };

        (void) psp_init();

        stat = tfserver_main(1, tfserver_main_argv);
        if (stat != S_OKAY)
        {
            cerr << "Error starting tfserver (" << stat << ")" << endl;
            psp_term();
            exit(EXIT_FAILURE);
        }
    }

    return pid;
}

/* Stop a process */
static void stop_process(
    pid_t pid)
{
    int ret;
    int stat;

    /* Kill the tfserver process */
    ret = kill(pid, SIGINT);
    if (ret == -1)
    {
        cerr << "Error killing tfserver (" << strerror(errno) << ")" << endl;
        exit(EXIT_FAILURE);
    }

    /* Wait for the tfserver process to terminate */
    do
    {
        waitpid(pid, &stat, 0);
    } while (!WIFEXITED(stat) && !WIFSIGNALED(stat));
}

/* Deleter thread */
static void *DeleterThreadFcn(
    void *arg)
{
    obj15    obj;
    int32_t *pn_ret;

    pn_ret = new int32_t;
    if (pn_ret == NULL)
    {
        cerr << "Error allocating memory..." << endl;
        return NULL;
    }

    try
    {
        obj.Initialize();
        obj.DeleteData();
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") in DeleterThread" << endl;
        *pn_ret = e.error();
        return pn_ret;        
    }

    *pn_ret = S_OKAY;
    return pn_ret;    
}

/* Inserter thread */
static void *InserterThreadFcn(
    void *arg)
{
    obj15    obj;
    int32_t *pn_ret;

    pn_ret = new int32_t;
    if (pn_ret == NULL)
    {
        cerr << "Error allocating memory..." << endl;
        return NULL;
    }

    try
    {
        obj.Initialize();
        obj.InsertData();
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") in InserterThread" << endl;
        *pn_ret = e.error();
        return pn_ret;
    }

    *pn_ret = S_OKAY;
    return pn_ret;
}

/* Reader thread */
static void *ReaderThreadFcn(
    void *arg)
{
    obj15    obj;
    int32_t *pn_ret;
 
    pn_ret = new int32_t;
    if (pn_ret == NULL)
    {
        cerr << "Error allocating memory..." << endl;
        return NULL;
    }

    try
    {
        obj.Initialize();
        obj.DisplayData();
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") in ReaderThread" << endl;
        *pn_ret = e.error();
        return pn_ret;
    }

    *pn_ret = S_OKAY;
    return pn_ret;
}

int32_t main(void)
{
    TFS_HANDLE  hTFS;
    pthread_t   inserter_tid;
    pthread_t   deleter_tid;
    pthread_t   reader_tid;
    int32_t    *ret_i;
    int32_t    *ret_d;
    int32_t    *ret_r;
    pid_t       pid_tfs;
    obj15       obj;

    pid_tfs = start_tfs_process();
    if (pid_tfs > 0)
    {
        /* Sleep to give the tfserver time to initialize and start listening */
        sleep(5);

        /* Initialize the application to use the external TFS (RPC) that we 
         * started */
        ::d_tfsinitEx(NULL, NULL, &hTFS, TFS_TYPE_RPC);

        try
        {
            obj.Initialize(1);
        }
        catch (rdm_exception& e)
        {
            cerr << "Error " << e.error() << " (" << e.what() << ")" << endl;
            return EXIT_FAILURE;
        }

        cout << endl << "Starting Threads..." << endl;

        (void) ::pthread_create(&inserter_tid, NULL, &InserterThreadFcn, NULL);
        (void) ::pthread_create(&deleter_tid, NULL, &DeleterThreadFcn, NULL);
        (void) ::pthread_create(&reader_tid, NULL, &ReaderThreadFcn, NULL);

        cout << "Press <Enter> when you are ready to terminate them . . ." << endl;
        getchar();

        /* Tell the threads to quit */    
        quit = 1;

        /* Wait for the reader thread */
        ::pthread_join(reader_tid, (void **) &ret_r);
        cout << "Reader thread finished" << endl;
        free(ret_r);

        /* Wait for the deleter thread */
        ::pthread_join(deleter_tid, (void **) &ret_d);
        cout << "Deleter thread finished" << endl;
        free(ret_d);

        /* Wait for the inserter thread */
        ::pthread_join(inserter_tid, (void **) &ret_i);
        cout << "Inserter thread finished" << endl;
        free(ret_i);

        obj.Cleanup();
        ::d_tfsterm(hTFS);

        stop_process(pid_tfs);

        cout << endl << "Press <Enter> to continue . . .";
        getchar();
    }

    return EXIT_SUCCESS;
}
